# -*- coding: utf-8 -*-
"""
test lazy-regression-tests
"""

import sys
import os
import unittest
import json
import codecs
from collections import namedtuple

import re
import shutil
import difflib

pyver = sys.version_info.major


try:
    import unittest.mock as mock
except (ImportError,) as ei:
    import mock  # python 2?

try:
    from bs4 import BeautifulSoup as bs
except (ImportError,) as e:
    try:
        from BeautifulSoup import BeautifulSoup as bs
    except (Exception,) as e2:
        bs = None


import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)
from traceback import print_exc as xp

# from lib.utils import set_cpdb, set_rpdb, ppp, debugObject
# from lib.utils import fill_template, Subber, RescueDict

from lazy_regression_tests._baseutils import (
    set_cpdb,
    set_rpdb,
    ppp,
    debugObject,
    fill_template,
    Subber,
    RescueDict,
)

from lazy_regression_tests.filters import CSSRemoveFilter


import pdb


def cpdb(**kwds):  # pragma : no cover
    if cpdb.enabled == "once":
        cpdb.enabled = False  # type: ignore
        return True
    return cpdb.enabled


cpdb.enabled = False  # type: ignore


def rpdb():  # pragma : no cover
    return rpdb.enabled


rpdb.enabled = False  # type: ignore
#

from lazy_regression_tests.core import (
    lzrt_default_t_basename,
    lzrt_default_t_subdir,
    LazyMixin,
    LazyIOErrorCodes,
    OnAssertionError,
    DiffFormatter,
    lzrt_default_t_basename,
)

from lazy_regression_tests.utils import (
    DictionaryKeyFilter,
    _Filter,
    RemoveTextFilter,
    RegexRemoveSaver,
    KeepTextFilter,
)

##########################################################
# tests
##########################################################


lzrt_template_dirname = "/<someroot>/out/tests/%(subject)s/%(lazy_dirname_extras)s"
lzrt_template_dirname_got = "/<someroot>/out/tests/%(subject)s/%(lazy_dirname_extras)s"
lzrt_template_dirname_exp = "/<someroot>/wk/tests/%(subject)s/%(lazy_dirname_extras)s"
lzrt_template_basename = lzrt_default_t_basename

di_mock_env = dict(
    lzrt_template_dirname=lzrt_template_dirname,
    lzrt_template_dirname_got=lzrt_template_dirname_got,
    lzrt_template_dirname_exp=lzrt_template_dirname_exp,
    lzrt_template_basename=lzrt_template_basename,
)

di_mock_env_no_extras = dict(
    [(k, v.replace("/%(lazy_dirname_extras)s", "")) for k, v in di_mock_env.items()]
)

module_ = "builtins"
module_ = module_ if module_ in sys.modules else "__builtin__"
funcpath_open = "%s.open" % module_


class Foo(object):
    pass


if __name__ == "__main__":
    lazy_filename = os.path.splitext(os.path.basename(__file__))[0]
else:
    try:
        logger.error("Foo.__module__:%s:" % (Foo.__module__))
        lazy_filename = Foo.__module__.split(".")[-1]
        # lazy_filename = "xxx"
        # lazy_filename = os.path.splitext(os.path.basename(__file__))[0]
        logger.error(Foo.__module__)
    except (NameError,) as e:
        logger.error(debugObject(globals(), "globals"))
        logger.error(debugObject(dir(Foo), "Foo"))
        raise


def debug_env(self):
    ppp(self.lazy_environ, "lazyenv")


class LazyMixinBasic(LazyMixin):

    """
    these are for not-live tests.



    """

    debug_env = debug_env

    lazy_filename = lazy_filename

    tmp_formatted_data = None

    def _lazy_write(self, fnp, formatted_data):
        self.tmp_formatted_data = formatted_data


class Test_001_Configuration(LazyMixinBasic, unittest.TestCase):

    # old_os_environ = os.environ
    # try:
    #     os.environ = {}
    #     self.assertRaises(OSError, my_function)
    # finally:
    #     os.environ = old_os_environ

    def test_001_missing_configuration(self):
        try:
            try:
                ante = os.environ
                os.environ = {}
                try:
                    tdirname = self._lazy_get_t_dirname("exp")
                except (ValueError,) as e:
                    message = str(e)
                    self.assertTrue("lzrt_template_dirname" in message)

                tbasename = self._lazy_get_t_basename("exp")
                # should default to default
                self.assertEqual(lzrt_default_t_basename, tbasename)

            finally:
                os.environ = ante

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise


@mock.patch.dict(os.environ, di_mock_env)
class Test_Basics(LazyMixinBasic, unittest.TestCase):
    def test_extensions(self):
        try:
            fnp_root = self.lazy_fnp_exp_root()

            got = self._lazy_add_extension(fnp_root, "html")
            exp = "%s.html" % (fnp_root)
            self.assertEqual(exp, got)

            suffix = "test"
            exp2 = "%s.%s.html" % (fnp_root, suffix)
            got2 = self._lazy_add_extension(fnp_root, "html", suffix=suffix)
            self.assertEqual(exp2, got2)

            got3 = self._lazy_add_extension(fnp_root)
            self.assertEqual(fnp_root, got3)
        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    def test_format_data(self):
        self.assertEqual(
            "{}", self.lazy_format_dict({}), "lazy_format_dict returns json.string"
        )
        self.assertEqual("{}", self.lazy_format_data({}))
        self.assertEqual("{}", self.lazy_format_data("{}"))

    def test_format_unicode_data(self):

        data = u"foo"
        self.assertEqual(data, self.lazy_format_data(data))


class TestBasic(LazyMixinBasic, unittest.TestCase):
    def lazy_write_ioerror(self, *args, **kwds):
        logger.info("%s.lazy_write_ioerror" % (self))
        pass

    @mock.patch.dict(os.environ, di_mock_env)
    def test_001_equal_string(self):
        """sanity check - are we ok if exp == got"""

        try:
            exp = got = "foo"

            if pyver == 3:
                exp = str.encode(exp)
                got = str.encode(got)

            with mock.patch(funcpath_open, mock.mock_open(read_data=exp)):
                self.assertLazy(got, onIOError=LazyIOErrorCodes.pass_missing)

            with mock.patch(funcpath_open, mock.mock_open(read_data=exp)):
                self.assertLazy(got, ".txt")

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_mock_env)
    def test_002_filename(self):

        if __name__ == "__main__":
            name = os.path.splitext(os.path.basename(__file__))[0]
        else:
            name = self.__module__.split(".")[-1]

        print("self.__module__:%s" % (self.__module__))
        print("__file__:%s:" % (__file__))
        print("name:%s:" % (name))

        exp = got = "foo"

        if pyver == 3:
            exp = str.encode(exp)
            got = str.encode(got)

        with mock.patch(funcpath_open, mock.mock_open(read_data=exp)):
            self.assertLazy(got)

        print("fnp_got:%s:" % (self.lazytemp.fnp_got))

        msg = "name:%s:not in fnp_got:%s" % (name, self.lazytemp.fnp_got)
        self.assertTrue(name in self.lazytemp.fnp_got, msg)

    @mock.patch.dict(os.environ, di_mock_env)
    def test_003_suffix(self):

        got = "something"

        if pyver == 3:
            got = str.encode(got)

        suffix = "mysuffix"
        with mock.patch(funcpath_open, mock.mock_open(read_data=got)):
            self.assertLazy(got, "txt", suffix=suffix)

        msg = "suffix:%s:not in fnp_got:%s" % (suffix, self.lazytemp.fnp_got)
        self.assertTrue(suffix in self.lazytemp.fnp_got, msg)

    def test_004_attribute_names(self):
        try:
            mixin = LazyMixin()
            attrnames = dir(mixin)
            logger.info("dir(mixin):%s" % (attrnames))

            tolerated = ["verbose", "assertLazy"]

            prefix = "lazy"
            bad = []
            for attrname in attrnames:
                if attrname.startswith("__") and attrname.endswith("__"):
                    continue

                if attrname in tolerated:
                    continue

                try:
                    msg = "%s. should have started with _lazy or lazy" % attrname
                    attrname2 = attrname.lstrip("_")
                    self.assertTrue(attrname2.startswith(prefix), msg)
                except (AssertionError,) as e:
                    bad.append(attrname)

            bad.sort()
            bad and self.fail("naming convention:%s" % "\n  ".join(bad))

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_mock_env)
    def test_005_fnp_exp_root(self):
        try:
            exp = got = None
            exp = os.path.join(
                lzrt_template_dirname_exp
                % {"subject": "exp", "lazy_dirname_extras": ""},
                "test_lazymixin.%s.%s"
                % (self.__class__.__name__, self._testMethodName),
            )

            got = self.lazy_fnp_exp_root()
            msg = "\nexp:%s\n<>\ngot:%s" % (exp, got)
            self.assertEqual(exp, got, msg)

        except (Exception,) as e:  # pragma : no cover
            ppp(os.environ, "os.environ")
            ppp(dict(exp=exp, got=got, lazy_filename=self.lazy_filename))

            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_mock_env_no_extras)
    def test_006_fnp_got_root(self):
        try:
            exp = got = None
            exp = os.path.join(
                lzrt_template_dirname_got
                % {"subject": "got", "lazy_dirname_extras": ""},
                "test_lazymixin.%s.%s"
                % (self.__class__.__name__, self._testMethodName),
            )

            got = self.lazy_fnp_got_root()
            msg = "\nexp:%s\n<>\ngot:%s" % (exp, got)
            self.assertEqual(exp, got, msg)

        except (Exception,) as e:  # pragma : no cover
            ppp(os.environ, "os.environ")
            ppp(dict(exp=exp, got=got))
            if cpdb():
                pdb.set_trace()
            raise


lorem = """
<html><head><title></title></head><body>
<p>Lorem ipsum dolor sit amet, consectetur 
adipisicing elit. Consequatur Quidem, sint incidunt?</p></body></html>
"""


class Test_IOError_Handling(LazyMixin, unittest.TestCase):

    lazy_filename = lazy_filename

    debug_env = debug_env

    def _lazy_write(self, fnp, formatted_data):
        pass

    def lazy_format_html(self, data):
        self.formatted_by = "lazy_format_html"
        self.data = data

    data = formatted_by = handled_by = None

    def lazy_write_assertionerror(self, fnp, formatted_data, message, exc):
        try:
            self.handled_by = LazyIOErrorCodes.assertion
            self.assertEqual("IOError(%s)" % (fnp), formatted_data)
        except (AssertionError,) as e:
            raise
        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    def lazy_write_ioerror(self, fnp, formatted_data, message, exc):
        self.handled_by = "lazy_write_ioerror"
        raise IOError()

    di = di_mock_env.copy()
    di.update(lzrt_directive=LazyIOErrorCodes.assertion)

    @mock.patch.dict(os.environ, di.copy())
    def test_write_assertionerror(self):

        got = exp = "wont be found"
        try:

            with mock.patch(
                funcpath_open, mock.mock_open(read_data=exp)
            ) as mocked_open:

                self.lazy_environ["lzrt_directive"] = LazyIOErrorCodes.assertion

                mocked_open.side_effect = IOError("fake IOError")
                try:
                    self.assertLazy(got)
                except (AssertionError,) as e:
                    self.assertEqual(self.handled_by, LazyIOErrorCodes.assertion)
                except (Exception,) as e:  # pragma : no cover
                    # pdb.set_trace()
                    if rpdb():  # pragma : no cover
                        pdb.set_trace()
                    self.fail("should have AssertionError. got Exception:%s" % str(e))

                else:
                    self.fail("should have AssertionError")

        finally:
            self.lazy_environ.clear()

    @mock.patch.dict(os.environ, di_mock_env)
    def test_write_ioerror(self):

        got = exp = "wont be found"

        try:
            self.lazy_environ["lzrt_directive"] = LazyIOErrorCodes.ioerror

            with mock.patch(
                funcpath_open, mock.mock_open(read_data=exp)
            ) as mocked_open:

                mocked_open.side_effect = IOError("fake IOError")
                try:
                    self.assertLazy(got)
                except (IOError,) as e:
                    self.assertEqual(self.handled_by, "lazy_write_ioerror")

                except (AssertionError,) as e:
                    if (
                        self.lazy_environ.get("lzrt_directive")
                        == LazyIOErrorCodes.assertion
                    ):
                        self.assertEqual(self.handled_by, LazyIOErrorCodes.assertion)
                    else:
                        raise

                except (Exception,) as e:  # pragma : no cover
                    raise
                else:
                    self.fail("should have IOError")

        except (AssertionError, IOError) as e:
            raise

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise
        finally:
            self.lazy_environ.clear()


class LazyMixin_DirRdbname(LazyMixinBasic):
    """put `rdbname` attribute, if found, in the directory hieararchy"""

    lazy_dirname_extras = "rdbname"


def debug_expgot(exp, got, testee=None):

    li = ["\n"]
    if testee:
        li.append(str(testee))
    li.append("exp:%s:" % (exp))
    li.append("got:%s:" % (got))

    logger.info("\n  ".join(li))


class Test_DirRdbname(LazyMixin_DirRdbname, unittest.TestCase):
    @mock.patch.dict(os.environ, di_mock_env)
    def test_fnp_exp_root(self):
        try:
            self.rdbname = "mydb"

            exp = got = None
            exp = os.path.join(
                lzrt_template_dirname_exp
                % {"subject": "exp", "lazy_dirname_extras": self.rdbname},
                "test_lazymixin.%s.%s"
                % (self.__class__.__name__, self._testMethodName),
            )

            got = self.lazy_fnp_exp_root()
            msg = "\nexp:%s\n<>\ngot:%s" % (exp, got)
            self.assertEqual(exp, got, msg)

        except (Exception,) as e:  # pragma : no cover
            ppp(os.environ, "os.environ")
            debug_expgot(exp, got)

            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_mock_env)
    def test_fnp_exp_root_missing_rdbname(self):
        try:
            # self.rdbname = "mydb"

            exp = got = None
            exp = os.path.join(
                lzrt_template_dirname_exp
                % {"subject": "exp", "lazy_dirname_extras": ""},
                "test_lazymixin.%s.%s"
                % (self.__class__.__name__, self._testMethodName),
            )

            got = self.lazy_fnp_exp_root()
            msg = "\nexp:%s\n<>\ngot:%s" % (exp, got)
            self.assertEqual(exp, got, msg)

            if self.verbose:
                debug_expgot(exp, got, self)

        except (Exception,) as e:  # pragma : no cover
            ppp(os.environ, "os.environ")
            debug_expgot(exp, got)

            if cpdb():
                pdb.set_trace()
            raise


# di = dict(lzrt_template_dirname="/Users/jluc/kds2/out/tests/LazyMixin/%(subject)s")

livetests_dir = os.environ.get("lzrt_livetests_dir", "")

has_directory_to_write_to = os.path.isdir(livetests_dir)

di_livetest = {}

NO_TESTWRITES_MSG = """skipping Live Tests. no writeable test directory provided in environment variable $lzrt_livetests_dir """

if has_directory_to_write_to:

    # flush the tmp directory
    dirtgt = os.path.join(livetests_dir, "tmp")

    if os.path.isdir(dirtgt):
        shutil.rmtree(dirtgt)

    lzrt_template_dirname = os.path.join(dirtgt, lzrt_default_t_subdir)

    live_seed = dict(
        lzrt_template_dirname=lzrt_template_dirname,
        lzrt_template_basename=lzrt_default_t_basename,
        lzrt_template_dirname_exp=lzrt_template_dirname,
        lzrt_template_dirname_got=lzrt_template_dirname,
    )

    # pdb.set_trace()
    ppp(live_seed, "live_seed")


class BaseHtmlFilter(LazyMixinBasic, unittest.TestCase):

    patre_manual_remove = re.compile("csrfmiddlewaretoken|var\ssettings|nodiff")

    _li_remove = [
        RegexRemoveSaver("var\ssettings\s=\s", hitname="settings"),
        re.compile("var\scsrfmiddlewaretoken\s=\s"),
    ]

    def format_exp(self):
        exp = []
        for line in self.data.split("\n"):
            if not self.patre_manual_remove.search(line):
                exp.append(line)

        res = "\n".join(exp)

        self.exp = self.mock_exp = bs(res).prettify()
        if pyver == 3:
            self.mock_exp = str.encode(self.exp)

    def setUp(self):

        try:
            if self.__class__ == BaseHtmlFilter:
                return

            super(BaseHtmlFilter, self).setUp()

            li_remove = self._li_remove + getattr(self, "li_remove", [])

            self.lazy_filter_html = RemoveTextFilter(
                li_remove, f_notify=self.lazy_filter_notify
            )

            self.format_exp()

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    di = di_livetest.copy()

    @mock.patch.dict(os.environ, di)
    def test(self):
        try:
            if self.__class__ == BaseHtmlFilter:
                return

            with mock.patch(funcpath_open, mock.mock_open(read_data=self.mock_exp)):
                if rpdb():  # pragma : no cover
                    pdb.set_trace()
                self.assertLazy(self.data, extension="html")
        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise


class TestCssFilter(BaseHtmlFilter):
    data = """
<script>
var csrfmiddlewaretoken = 'wTNDVhWQHWzbf0Yb7mWo7PG03SgE9rpWfNXD3ZpbPm9IaZXAs3DuBUbOzI8oFutW';
var settings = {"li_user_message": []};
</script>
<div class="row">
    <div>keep ante</div>
    <div class="nodiff">remove this</div>
    <div>keep post</div>
</div>

    """

    li_remove = [CSSRemoveFilter(".nodiff")]


@unittest.skipUnless(has_directory_to_write_to, NO_TESTWRITES_MSG)
class TestLive(LazyMixin, unittest.TestCase):

    debug_env = debug_env

    def setUp(self):
        # raise NotImplementedError("%s.setUp" % (self))
        self.old_seed = self.lazy_environ.seed

        self.lazy_environ.seed = live_seed
        self.lazy_environ.clear()

    def tearDown(self):
        self.lazy_environ.seed = self.old_seed
        # raise NotImplementedError("%s.tearDown" % (self))
        self.lazy_environ.clear()

    lazy_filename = lazy_filename

    lazy_message_formatter = DiffFormatter()

    # this will be used, if found, in the directory names.
    lazy_dirname_extras = "rdbname"

    @mock.patch.dict(os.environ, di_livetest)
    def test_001_basic(self):
        got = "something"
        # this will warn only first time, because exp does not exist
        self.assertLazy(got, onIOError=LazyIOErrorCodes.pass_missing)

        # and will pass the 2nd because it was just created
        self.assertLazy(got)

        # and now it should fail
        try:
            got2 = "different"
            self.assertLazy(got2)
            self.fail("should have gotten %s<>%s AssertionError" % (got, got2))
        except (AssertionError,) as e:
            self.assertTrue(got in str(e))
            self.assertTrue(got2 in str(e))
        except (Exception,) as e:  # pragma : no cover
            self.fail("should have gotten %s<>%s AssertionError" % (got, got2))

    @mock.patch.dict(os.environ, di_livetest)
    def test_002_html(self):
        data = lorem

        self.assertLazy(data, "html", onIOError=LazyIOErrorCodes.pass_missing)

        # and will pass the 2nd because it was just created
        self.assertLazy(data, "html")

    @mock.patch.dict(os.environ, di_livetest)
    def test_003_suffix(self):
        got1 = "<div><span>got1</span></div>"
        got2 = "<div><span>got2</span></div>"

        self.assertLazy(
            got1, "html", onIOError=LazyIOErrorCodes.pass_missing, suffix="suffix1"
        )
        self.assertLazy(
            got2, "html", onIOError=LazyIOErrorCodes.pass_missing, suffix="suffix2"
        )

        self.assertLazy(got1, "html", suffix="suffix1")
        self.assertLazy(got2, "html", suffix="suffix2")

    di = di_livetest.copy()
    di.update(lzrt_directive=OnAssertionError.baseline)

    @mock.patch.dict(os.environ, di)
    def test_004_baseline(self):

        got = "something"
        got2 = "different"

        # OnAssertionError.baseline suppresses everything...
        self.assertLazy(got)

        with open(self.lazytemp.fnp_exp) as fi:
            self.assertEqual(got, fi.read())

        self.assertLazy(got2)

        # now check that exp == got2
        with open(self.lazytemp.fnp_exp) as fi:
            self.assertEqual(got2, fi.read())

    di = di_livetest.copy()
    di.update(lzrt_directive="skip")

    @mock.patch.dict(os.environ, di)
    def test_0041_skip(self):

        got = "something"
        got2 = "different"

        # OnAssertionError.baseline suppresses everything...
        self.assertLazy(got)

    @mock.patch.dict(os.environ, di_livetest)
    def test_005_rdbname_in_directory(self):
        got = "something"
        got2 = "different"
        db1 = "db1"
        db2 = "db2"

        self.rdbname = db1
        self.assertLazy(got, onIOError=LazyIOErrorCodes.pass_missing)

        self.assertTrue(db1 in self.lazytemp.fnp_exp)
        self.assertTrue(db1 in self.lazytemp.fnp_got)

        # and this should fail with an IOError because it's not the
        # same rdbname so File.exp is missing.
        self.rdbname = db2
        try:
            self.assertLazy(got)
            self.fail("should have gotten %s<>%s AssertionError" % (got, got2))
        except (IOError,) as e:
            logger.info("got my IOError")
            self.assertTrue(db2 in str(e))
        except (Exception,) as e:  # pragma : no cover
            self.fail("should have gotten %s<>%s AssertionError" % (got, got2))

        # now this should work
        self.rdbname = db1
        self.assertLazy(got)

    @mock.patch.dict(os.environ, di_livetest)
    def test_006_dict(self):
        try:
            data = dict(a=1, b=2, c=3)
            self.assertLazy(data, "json", onIOError=LazyIOErrorCodes.pass_missing)
            data.update(d=4)
            try:
                self.assertLazy(data, "json")
            except (AssertionError,) as e:
                # pdb.set_trace()
                pass
                found = False
                lines = [line for line in str(e).splitlines() if line.startswith("+")]
                # pdb.set_trace()
                for line in lines:
                    if '"d"' in line:
                        found = True
                        break
                self.assertTrue(
                    found, "lines:%s, should have included the extra d=4 item" % (lines)
                )

            else:
                self.fail("should have had an AssertionError")

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_livetest)
    def test_007_filter(self):

        try:
            data = """
            blah blah
            skip
            dude
            """

            ignore = "skip"

            self.lazy_filter_html = RemoveTextFilter(
                [ignore], f_notify=self.lazy_filter_notify
            )

            self.assertLazy(data, "html", onIOError=LazyIOErrorCodes.pass_missing)

            with open(self.lazytemp.fnp_exp) as fi:
                written = fi.read()
            self.assertFalse(ignore in written, written)

            # check what the filter stripped out
            # self.assertTrue(ignore in list(self.lazytemp.filterhits.values())[0].found)

            data = data.replace(ignore, "%s this" % (ignore))

            self.assertLazy(data, "html")

            self.lazy_filter_txt = KeepTextFilter(
                ["a", "b", "c"], f_notify=self.lazy_filter_notify
            )

            data = """a
            a
            b
            d
            """

            self.assertLazy(data, "txt", onIOError=LazyIOErrorCodes.pass_missing)

            with open(self.lazytemp.fnp_exp) as fi:
                written = fi.read()
            self.assertFalse("d" in written)

            data = """a
            a
            b
            e
            """
            self.assertLazy(data, "txt")

            filter_ = KeepTextFilter(["a", "b"])

            self.assertLazy(
                data, "text", onIOError=LazyIOErrorCodes.pass_missing, filter_=filter_
            )
            with open(self.lazytemp.fnp_exp) as fi:
                written = fi.read()

            exp = """a
            a
            b
            """

            self.assertEqual(written.strip(), exp.strip())
        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_livetest)
    def test_0071_filter_regex(self):

        try:
            data = """
var debug_chart = 2;

// settings are set up by <settings object>.resolve_as_json_on_context;
var settings = {"tags": [{"disable_override": true, "checked": null, "description": "manage security", "choices_value2": null, "hide_message": true, "allow_entry_ivalue1": false, "system_write": true, "allow_entry_cvalue2": false, "ivalue1": null, "tagtype": "M_SEC", "cvalue2": null, "override": null, "message_value": ""}, {"disable_override": true, "checked": true, "description": "User count", "choices_value2": null, "hide_message": false, "allow_entry_ivalue1": false, "system_write": true, "allow_entry_cvalue2": false, "ivalue1": 139, "tagtype": "M_UCN", "cvalue2": " ", "override": null, "message_value": "139"}], "portalname": "EMPLOYEE", "li_rdbname": ["pgfin92"], "is_superuser": false, "tracker": {"projectname": "", "main": {"objectvalue1": "PeopleSoft User", "projectname": "", "objecttype": -1}, "psprsmdefn_detail": {"objectvalue1": "", "objectvalue2": "", "objectvalue3": "", "projectname": "", "objecttype": 55}}, "li_user_message": [], "objectvalue01": "PeopleSoft User", "ROLENAME": "PeopleSoft User", "rdbname": "hcm91dmo"};
var csrfmiddlewaretoken = 'wTNDVhWQHWzbf0Yb7mWo7PG03SgE9rpWfNXD3ZpbPm9IaZXAs3DuBUbOzI8oFutW';
var csrf_token = 'wTNDVhWQHWzbf0Yb7mWo7PG03SgE9rpWfNXD3ZpbPm9IaZXAs3DuBUbOzI8oFutW';

    <script>
    
    // settings are set up by <ViewManager>.calc_json_settings()
    var settings = {"li_user_message": [], "rdbname_0": "hcm91dmo", "rdbname_1": "pgfin92", "objectvalue01": "FEDTBHADMN1", "objectvalue11": "AMARTIN", "xdb_mode": true, "rdbname_r": "pgfin92", "rdbname_l": "hcm91dmo"};
    
    </script>


// require( ["pssecurity/detail_require"], function(module){
// module.init();
// });
"""

            li_remove = [
                # re.compile("var\ssettings\s=\s"),
                # RegexRemoveSaver saves the hit in list temp.filterhis[hitname]
                RegexRemoveSaver("var\ssettings\s=\s", hitname="settings"),
                re.compile("var\scsrfmiddlewaretoken\s=\s"),
                re.compile("var\scsrf_token\s=\s"),
            ]

            self.lazy_filter_html = RemoveTextFilter(
                li_remove, f_notify=self.lazy_filter_notify
            )

            temp = self.assertLazy(
                data, "html", onIOError=LazyIOErrorCodes.pass_missing
            )

            self.assertTrue(
                self.lazytemp.filterhits["settings"][0].found.startswith("var settings")
            )

            with open(self.lazytemp.fnp_exp) as fi:
                written = fi.read()

            self.assertFalse("csrf" in written, written)
            self.assertFalse("var settings" in written, written)

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    @mock.patch.dict(os.environ, di_livetest)
    def test_010_unequal_string(self):

        try:

            exp = "foo"
            got = exp + ".unexpected"

            self.assertLazy(exp, onIOError=LazyIOErrorCodes.pass_missing)
            try:
                self.assertLazy(got)
            except (AssertionError,) as e:
                pass
            else:
                self.fail("should have failed")

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise


class TestThrottling(LazyMixin, unittest.TestCase):
    def setUp(self):
        self.lazy_message_formatter = DiffFormatter(maxlines=5)

    def test_it(self):
        exp = "\n".join([str(i) for i in range(0, 100)])
        got = "\n".join([str(i + 100) for i in range(0, 100)])

        message = self.lazy_message_formatter.format(exp, got)

        print(message)


class TestFilters(unittest.TestCase):

    matchers = ["csrf", "sysdate"]
    _data = dict(csrf="skip.csrf", sysdate="skip.sysdate", keep1="keep1", keep2="keep2")

    def setUp(self):
        self.watcher = _Filter(self.matchers)
        self.data = self._data.copy()

    def test_001_create(self):
        self.assertTrue(self.watcher)

    def test_002_scan(self):

        try:
            msg = None
            res = self.watcher.scan(self.data)

            for key in self.matchers:
                # if self.data.has_key(key):
                if key in self.data:
                    found = None
                    for f_ in res.finds:
                        if key == f_.found[0]:
                            found = True

                    msg = "did not find %s in res.finds:%s" % (key, res.finds)
                    self.assertTrue(found, msg)

        except (Exception,) as e:  # pragma : no cover
            logger.info(msg)
            logger.error(str(e))
            if cpdb():
                pdb.set_trace()
            raise

    def test_003_remove(self):

        try:
            msg = None
            self.watcher = DictionaryKeyFilter(self.matchers)
            # self.watcher.callback = worker.process
            exp = self.data.copy()

            res = self.watcher.scan(self.data)

            self.assertTrue(
                self.watcher.callback,
                "DictionaryKeyFilter.callback should self.worker.process",
            )

            for key in self.matchers:
                try:
                    del exp[key]
                except (KeyError,) as e:
                    pass

            self.assertEqual(exp, self.data)

        except (Exception,) as e:  # pragma : no cover
            logger.info(msg)
            logger.error(str(e))
            if cpdb():
                pdb.set_trace()
            raise


class TestFilters_Absent(TestFilters):

    matchers = ["xcsrf", "xsysdate"]
    _data = dict(csrf="skip.csrf", sysdate="skip.sysdate", keep1="keep1", keep2="keep2")


if __name__ == "__main__":

    set_cpdb(cpdb, remove=True)
    set_rpdb(rpdb, remove=True)

    sys.exit(unittest.main())



"""docstring"""
import unittest
import sys
import os
import logging
import argparse
import codecs
import hashlib

import glob
import re

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


from lazy_regression_tests._baseutils import debugObject, ppp, Dummy, set_cpdb, set_rpdb


from importlib import import_module

from traceback import print_exc as xp
import pdb


def cpdb():
    return cpdb.enabled


cpdb.enabled = False  # type: ignore


def rpdb():  # pragma : no cover
    return rpdb.enabled


rpdb.enabled = False  # type: ignore
if __name__ == "__main__":
    set_cpdb(cpdb, remove=True)
    set_rpdb(rpdb, remove=True)


class MainManager(object):
    """ manages batch"""

    di_opt = {}

    def __init__(self, options):

        try:
            self.options = options

            msg = debugObject(self, "\n\n MainManager.__init__:end")
            msg += debugObject(self.options, "\noptions")

            self.func_filter_factory = import_string(self.options.filter_builder)
            self.filter_ = self.func_filter_factory(onlyonce=True)

            if self.options.output_dir and not os.path.isdir(self.options.output_dir):
                raise ValueError(
                    "output_dir:%s does not exist" % (self.options.output_dir)
                )

            logging.info(msg)
            print(msg)

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    def process_input_directory(self, dir_in):
        try:
            globmask = getattr(self.func_filter_factory, "globmask", "*")

            li = glob.glob(os.path.join(dir_in, globmask))
            for fnp in li:
                self.reformat(fnp)

            if not self.options.write:
                raise NotImplementedError(
                    "need to implement output_dir on input dir.  for now only write-in-place is supported"
                )
        except (Exception,) as e:  # pragma : no cover
            print("\n\nli (first 5):")
            for fnp in li[0:5]:
                print(fnp)

            if cpdb():
                pdb.set_trace()
            raise

    def process(self):
        try:

            if self.options.input_ and os.path.isfile(self.options.input_):
                self.reformat(self.options.input_)
            elif self.options.input_ and os.path.isdir(self.options.input_):
                self.process_input_directory(self.options.input_)
            else:
                raise NotImplementedError()
        except (Exception,) as e:  # pragma : no cover
            ppp(self, self)
            ppp(self.options, "options")
            if cpdb():
                pdb.set_trace()
            raise

    def _hash(self, data):
        h = hashlib.sha1()
        try:
            h.update(data)
        except (UnicodeEncodeError,) as e:
            return None
        return h.digest()

    def reformat(self, fnp_i):
        try:

            with codecs.open(fnp_i, encoding="utf-8", errors="ignore") as fi:
                previous = fi.read()

            h_previous = self._hash(previous)

            refiltered = self.filter_(previous)

            h_refiltered = self._hash(refiltered)
            if h_previous is not None and h_previous == h_refiltered:
                logger.info("no changes for %s" % (fnp_i))
                return

            if self.options.output_dir:
                basename = os.path.basename(fnp_i)
                fnp_o = os.path.join(self.options.output_dir, basename)

                logger.info("\ndiff %s %s" % (fnp_i, fnp_o))
            else:
                fnp_o = fnp_i

            with codecs.open(fnp_o, encoding="utf-8", errors="ignore", mode="w") as fo:
                fo.write(refiltered)

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    @classmethod
    def getOptParser(cls):

        parser = argparse.ArgumentParser()

        dest = "filter_builder"
        parser.add_argument(
            dest=dest,
            action="store",  # store_true, store_false
            help="python import path to a function returning a lazy_regression_tests.utils.KeepTextFilter/RemoveTextFilter",
        )

        dest = "input_"
        parser.add_argument(
            dest=dest,
            action="store",  # store_true, store_false
            help="%s dump file needing formatting" % (dest),
        )

        dest = "output_dir"
        parser.add_argument(
            "--" + dest,
            action="store",  # store_true, store_false
            help="%s output directory, if not in-place" % (dest),
        )

        dest = "write"
        default = False
        # choices=["a","b","c"]
        parser.add_argument(
            "--" + dest,
            default=default,
            action="store_true",
            help="%s [%s]" % (dest, default),
        )

        return parser


########################
# shamelessly copied from django.utils.module_loading
########################


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """

    try:

        try:
            module_path, class_name = dotted_path.rsplit(".", 1)
        except ValueError:
            msg = "%s doesn't look like a module path" % dotted_path
            raise ImportError

        module = import_module(module_path)

        try:
            return getattr(module, class_name)
        except AttributeError:
            msg = 'Module "%s" does not define a "%s" attribute/class' % (
                module_path,
                class_name,
            )
            raise ImportError

    except (Exception,) as e:  # pragma : no cover

        if cpdb():
            ppp(locals(), "\nlocals")
            print("sys.path:")
            for pa in sys.path:
                print(pa)

            pdb.set_trace()
        raise


if __name__ == "__main__":

    parser = MainManager.getOptParser()
    options = parser.parse_args()
    mgr = MainManager(options)
    mgr.process()



# -*- coding: utf-8 -*-
"""Mixin to support regression testing lazily:

       given a some received data (referred as `got`)
       does it match what was seen, and saved, the last time?

       Default naming: <module>.<classname>.<function>.<extension>

          but you can specify instance attributes to use in
          either the directory or file names

       A typical test write might save data as follows:

       For each `got` data passed to assertLazy, the mixin will

       1.  save the `got` to the `got` directory branch.
       2.  load the matching `exp` file 

          `exp` files that do not exist get created with `got` data.
          this will be used as `exp` the next time around.
    
          an error will be thrown, but setting onIOerror = "pass_missing" will 
          suppress that.
        
       3.  then `assertEqual` them.


/Users/jluc/kds2/wk/issues/nfb022.utils/issues/013.utmixin/temp/tmp
├── exp
│   ├── db1
│   │   └── test_uthelper_mixin.TestLive.test_rdbname_in_directory
│   ├── db2
│   │   └── test_uthelper_mixin.TestLive.test_rdbname_in_directory
│   ├── test_uthelper_mixin.TestLive.test_basic
│   ├── test_uthelper_mixin.TestLive.test_html.html
│   └── test_uthelper_mixin.TestLive.test_ignore
└── got
    ├── db1
    │   └── test_uthelper_mixin.TestLive.test_rdbname_in_directory
    ├── db2
    │   └── test_uthelper_mixin.TestLive.test_rdbname_in_directory
    ├── test_uthelper_mixin.TestLive.test_basic
    ├── test_uthelper_mixin.TestLive.test_html.html
    └── test_uthelper_mixin.TestLive.test_ignore
"""

#######################################################
# Typing
#######################################################

from typing import Any

###################################################################

import sys
import os
import unittest
import json
import codecs

import re
import shutil


###################################################################


try:
    from bs4 import BeautifulSoup as bs
except (ImportError,) as e:
    try:
        from BeautifulSoup import BeautifulSoup as bs
    except (Exception,) as e2:
        bs = None

import logging


from yaml import dump as ydump, safe_load as yload


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)
from traceback import print_exc as xp

try:
    from lazy_regression_tests.utils import (
        DiffFormatter,
        replace,
        MediatedEnvironDict,
        Found,
        basestring_,
        unicode_,
        debugObject,
        fill_template,
        Subber,
        RescueDict,
        ppp,
    )
except (ImportError,) as e:
    # not sure if needed
    # from .utils import DiffFormatter
    raise

import pdb


def cpdb():
    return cpdb.enabled


cpdb.enabled = False  # type: ignore


def rpdb():  # pragma : no cover
    return rpdb.enabled


rpdb.enabled = False  # type: ignore
###################
# configuration
#####################

MISSING_ENV_T_DIRNAME = """
******************************************************************************
lazy-regression-tests:  Invalid configuration 
The template for the directory name could not be found 
for subject `%(subject)s`.
Try setting either `$%(specific)s` or the generic `$%(generic)s`.
This needs to point to a user-writeable directory of your choice.
******************************************************************************
"""


#########################
# environment variable names
#######################
env_directive = "lzrt_directive"
env_t_dirname = "lzrt_template_dirname"
env_t_basename = "lzrt_template_basename"


from collections import namedtuple

Choice = namedtuple("Choice", "code help")

opt_nodiff = "nodiff"


class DirectiveChoices(object):
    """what can go into environment variable `lzrt_directive"""

    baseline = Choice(
        "baseline",
        """establish a baseline behavior - all mismatches and IOError are ignored
and existing expecations are reset""",
    )
    skip = Choice("skip", """do not run lazy-regression-tests""")

    nodiff = Choice(opt_nodiff, """check equality, but don't run diff""")

    missing_pass = Choice(
        "missing_pass",
        """IOError on expectations will be ignored and treated as a match success
the formatted `got/received` value will be written to the expectation.
""",
    )
    assert_missing = Choice(
        "assert_missing",
        """IOError on expectations throw an AssertionError instead of an IOError.
the formatted `got/received` value will be written to the expectation.        
""",
    )

    @classmethod
    def output_help(cls, writer):

        writer("\n")
        writer("possible values in $lzrt_directive:\n")
        for k in dir(cls):
            v = getattr(cls, k)
            if isinstance(v, Choice):
                option = "'%s'" % (k)
                writer("  %-20s : %s\n" % (option, v.help))


class OnAssertionError(object):

    # use `baseline` when you want to reset the whole codeline to new expectations
    # can only be changed on the environment level, or via command line option
    # also implies onIOError.pass_missing
    baseline = DirectiveChoices.baseline.code

    # standard assertEqual behavior
    default = "error"

    # we are not running these tests
    ignore = DirectiveChoices.skip.code

    nodiff = DirectiveChoices.nodiff.code


class LazyIOErrorCodes(object):
    """what happens when read.exp throws an IOError?"""

    # default behavior: `got` => File.exp, but throw an IOError
    # differentiates data mismatch from missing `exp` files
    default = ioerror = "lazy_write_ioerror"

    # write `got` to File.exp, without throwing an exception.
    # useful when first running, but see also OnAssertionError.baseline mode
    pass_missing = "lazy_write_passmissing"

    # default behavior: `got` => File.exp, but throw an AssertionError instead
    assertion = "lazy_write_assertionerror"


SYSARG_BASELINE = "--lazy-%s" % OnAssertionError.baseline
SYSARG_IGNORE = "--lazy-%s" % OnAssertionError.ignore
SYSARG_PASS_MISSING = "--lazy-%s" % LazyIOErrorCodes.pass_missing.replace("_", "-")
SYSARG_NODIFF = "--lazy-%s" % OnAssertionError.nodiff


# allows per subject-environment lookups i.e. got may be put somewhere else than exp
t_env_dirname_subject = "lzrt_template_dirname_%(subject)s"

##############
# defaults
##############
lzrt_default_t_subdir = "%(subject)s/%(lazy_dirname_extras)s"
lzrt_default_t_basename = "%(filename)s %(classname)s %(_testMethodName)s %(lazy_basename_extras)s %(suffix)s %(extension)s"


###################
# utility functions and classes
####################


class _Control(object):
    """unifies environment and function arguments
       to determine handlers for IOError and AssertionError
       save in the LazyTemp results object as well.
    """

    def __init__(
        self, mixin: "LazyMixin", env: MediatedEnvironDict, onIOError_, **kwargs
    ):
        pass

        if not env or not env.acquired:
            env.acquire()

        directive = env.get(env_directive)

        self.skip = directive == OnAssertionError.ignore

        self.nodiff = directive == OnAssertionError.nodiff

        self.nest_testfilename = kwargs.get("nest_testfilename")

        self.lazy_filename = kwargs.get("lazy_filename")

        self.baseline = env.get(env_directive) == OnAssertionError.baseline
        if self.baseline:
            self.handler_io_error = mixin.lazy_write_passmissing
        else:
            funcname_handler_ioerror = (
                # specified in the assertLazy call?
                onIOError_
                # environment?
                or env.get(env_directive)
                # default value on instance
                or mixin.lazy_onIOError
            )
            self.handler_io_error = (
                getattr(mixin, funcname_handler_ioerror, None)
                or mixin.lazy_write_ioerror
            )


class LazyTemp(object):
    def __init__(self, control, env):
        self.fnp_exp = self.fnp_got = None
        self.env = env.copy()
        self.filterhits = {}
        self.control = control

    def notify(self, found, by_=None):
        try:
            # only save hits that have a finder.hitname

            hitname = getattr(getattr(found, "by", None), "hitname", None) or getattr(
                by_, "hitname", None
            )

            if hitname:
                li = self.filterhits.setdefault(hitname, [])
                li.append(found)

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise


ENV_PREFIX = "lzrt_"


PATRE_YAML_OBJECTSPEC = re.compile("!!python/.+$")


def yaml_to_dict(data, track_type=False):
    """take an arbitrary python object and transform it to dict form"""
    try:

        #
        str_yaml = ydump(data, default_flow_style=False)

        lines = []
        for line in str_yaml.split("\n"):

            hit = PATRE_YAML_OBJECTSPEC.search(line)
            if hit:
                line2 = PATRE_YAML_OBJECTSPEC.sub("", line).rstrip()
                lines.append(line2)

                # this isnt working yet...
                if track_type:
                    lines.append("""  _ytype : "%s" """ % line[hit.start() : hit.end()])

            else:
                lines.append(line)

        safe_yaml = "\n".join(lines)

        res = yload(safe_yaml)

        return res
    except (Exception,) as e:  # pragma : no cover
        if cpdb():
            pdb.set_trace()
        raise


class LazyMixin(object):
    """main class.  see `assertLazy`"""

    lazy_filename = "needs-setting-in-your-unitest-file"

    lazy_onIOError = LazyIOErrorCodes.default
    lazy_rescuedict = RescueDict(template="")

    lazy_environ = MediatedEnvironDict(filters=[ENV_PREFIX])

    # default for extra attributes.
    lazy_dirname_extras = ""
    lazy_basename_extras = ""

    lazytemp = None
    lazy_message_formatter = None

    lazy_message_formatter
    lazy_message_formatter = DiffFormatter()

    def lazy_filter_notify(self, found):
        self.lazytemp.notify(found)

    @classmethod
    def lazy_format_dict(cls, dict_, filter=None):
        return unicode_(
            json.dumps(dict_, sort_keys=True, indent=4, separators=(",", ":"))
        ).strip()

    @classmethod
    def lazy_format_dict2yaml(cls, dict_, filter=None):
        return ydump(dict_, default_flow_style=False)

    @property
    def verbose(self):
        res = sys.argv.count("-v")
        return res

    @classmethod
    def _lazy_get_t_dirname(cls, subject="", control=None):

        env_t_dirname_specific = fill_template(
            t_env_dirname_subject, {"subject": subject}
        )

        env = cls.lazy_environ
        if not env or not env.acquired:
            env.acquire()

        res = env.get(env_t_dirname_specific) or env.get(fill_template(env_t_dirname))

        if res is None:
            raise ValueError(
                MISSING_ENV_T_DIRNAME
                % dict(
                    subject=subject,
                    specific=env_t_dirname_specific,
                    generic=env_t_dirname,
                )
            )

        # nesting each test script under its own subdirectory
        nest_testfilename = getattr(control, "nest_testfilename", False)
        nest_testfilename = getattr(control, "lazy_filename", "")
        if nest_testfilename:
            res = os.path.join(res, control.lazy_filename)

        return res

    @classmethod
    def _lazy_get_t_basename(cls, subject):

        env = cls.lazy_environ
        if not env or not env.acquired:
            env.acquire()

        res = env.get(fill_template(env_t_basename)) or lzrt_default_t_basename
        return res

    def _lazy_write(self, fnp, formatted_data):

        dirname = os.path.dirname(fnp)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        if self.verbose:
            logger.info("%s._write:%s" % (self, fnp))

        with codecs.open(fnp, mode="w", encoding="utf-8", errors="ignore") as fo:
            fo.write(formatted_data)

    def lazy_write_assertionerror(self, fnp, formatted_data, message, exc=None):
        try:
            self._lazy_write(fnp, formatted_data)
            self.assertEqual(str(IOError(fnp)), formatted_data, message)
        except (AssertionError,) as e:
            if rpdb():  # pragma : no cover
                pdb.set_trace()
            raise
        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    def lazy_write_passmissing(self, fnp, formatted_data, message, exc=None):
        try:
            logger.warning("%s.suppressed IOError:%s" % (self, exc))
            self._lazy_write(fnp, formatted_data)
        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    def lazy_write_ioerror(self, fnp, formatted_data, message, exc):
        try:
            self._lazy_write(fnp, formatted_data)
            self.fail(str(exc))
            # raise exc
        except (IOError,) as e:
            if rpdb():  # pragma : no cover
                pdb.set_trace()
            raise
        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    def lazy_fnp_exp_root(self, control=None):
        return self._lazy_get_fnp_root(subject="exp", control=control)

    def lazy_fnp_got_root(self, control=None):
        return self._lazy_get_fnp_root(subject="got", control=control)

    def lazy_format_string(self, data, filter=None):

        if isinstance(data, unicode_):
            return data.strip()

        return unicode_(data, encoding="utf-8", errors="ignore").strip()

    def lazy_format_html(self, data, filter=None):
        # raise NotImplementedError("%s.lazy_format_html(%s)" % (self, locals()))

        data = self.lazy_format_string(data)
        if bs is None:
            logger.warning("BeautifulSoup is unavaiable")
            return data
        soup = bs(data)  # make BeautifulSoup

        if filter:
            soup = filter.format(soup, lazytemp=self.lazytemp)

        if rpdb():
            pdb.set_trace()

        return soup.prettify()  # prettify the html

    def lazy_format_json(self, data, filter=None):
        di = json.loads(data)
        return self.lazy_format_dict(di)

    def lazy_format_yaml(self, data, filter=None):
        try:
            di = yaml_to_dict(data)
            return self.lazy_format_dict2yaml(di)
        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    def lazy_format_data(self, data, extension="", filter=None):

        if isinstance(data, dict):
            res = self.lazy_format_dict(data, filter=filter)
            if isinstance(res, basestring_):
                res = res.strip()
            return res

        elif isinstance(data, bytes):
            data = data.decode("utf-8")
            if extension == "html":
                # track the cast
                data = (
                    "<!-- lazy_format_data.info:cast via bytes.decode(utf-8) -->\n%s"
                    % (data)
                )

            f = (
                getattr(self, "lazy_format_%s" % (extension.lower()), None)
                or self.lazy_format_string
            )
            return f(data).strip()

        elif isinstance(data, basestring_):

            f = (
                getattr(self, "lazy_format_%s" % (extension.lower()), None)
                or self.lazy_format_string
            )
            return f(data, filter=filter).strip()

        elif extension == "yaml":
            return self.lazy_format_yaml(data)
        else:
            di_diag = dict(
                data=str(data)[:100], datatype=type(data), extension=extension
            )

            raise NotImplementedError("%s.lazy_format_data(%s)" % (self, di_diag))
            return self.lazy_format_string(data).strip()

    def _lazy_get_fnp_root(self, subject, control=None):
        """get the root name, before extension and suffix"""

        subber = Subber(
            self,
            {
                "filename": self.lazy_filename,
                "subject": subject,
                "classname": self.__class__.__name__,
            },
            self.lazy_rescuedict,
        )

        # pdb.set_trace()

        # calculating the directory path
        t_dirname = self._lazy_get_t_dirname(subject=subject, control=control)
        _litd = t_dirname.split(os.path.sep)

        dirname_extras = getattr(self, "lazy_dirname_extras", "")
        if dirname_extras:
            # expand something like "foo, bar" into [..."%(foo)s", "%(bar)s"...]
            li_replace = ["%%(%s)s" % (attrname) for attrname in dirname_extras.split()]

            if "%(lazy_dirname_extras)s" in _litd:
                _litd = replace(_litd, "%(lazy_dirname_extras)s", li_replace)
            else:
                _litd.extend(li_replace)

        _lid = ["/"] + [fill_template(t_, subber) for t_ in _litd]

        dirname = os.path.join(*_lid)

        # calculating the filename
        t_basename = self._lazy_get_t_basename(subject)
        _litb = t_basename.split()
        _lib = [fill_template(t_, subber) for t_ in _litb]
        basename = ".".join([i_ for i_ in _lib if i_])

        basename = basename.replace(" ", "_")
        basename = basename.replace("/", "_")

        return os.path.join(dirname, basename)

    def _lazy_add_extension(self, fnp, extension="", suffix=""):
        """adds suffix, extension to the filename"""

        # don't add empty parts...
        li = [str(i_) for i_ in [fnp, suffix, extension] if i_]

        return ".".join(li)

    def assertLazy(
        self,
        got,
        extension="",
        suffix="",
        onIOError=None,
        message=None,
        filter_=None,
        formatter=None,
        f_notify=None,
        no_assert=False,
        nest_testfilename=True,
    ):
        """ check that result matches expectations saved previously.
        when the expectations file doesn't exist yet, it is created with received data
        but an error is raised unless suppressed by `onIOError`.

        :param got :  received data, to be checked against expectation
        :param extension :  used determining file names but also data formatting
        :param onIOError :  handler for when the expectation file can't be found
        :param message:  same as other assert's message parameter
        :param filter_: a filtering function f(string)=>string
        :param formatter: a function f(data, extension)=>string
        :param f_notify: if using the default filter and this is empty, reroutes notifications
                         into default filter
        """

        try:
            env = self.lazy_environ
            if not bool(self.lazy_environ):
                env.clear()
                env.acquire()

            control = _Control(
                self,
                env,
                onIOError,
                **{
                    "nest_testfilename": nest_testfilename,
                    "lazy_filename": self.lazy_filename,
                }
            )

            smsg = "%s.assertLazy:" % (self)

            tmp = self.lazytemp = LazyTemp(control, env)
            tmp.got = got

            if control.skip:
                if self.verbose:
                    logger.info("%s skipping lazy checks" % (smsg))
                return

            tmp.fnp_exp = fnp_exp = self._lazy_add_extension(
                self.lazy_fnp_exp_root(control=control), extension, suffix
            )

            # is there a filter for the extension?
            filter_ = filter_ or getattr(self, "lazy_filter_%s" % (extension), None)

            formatter = formatter or self.lazy_format_data
            if got:
                formatted_data = formatter(got, extension, filter=filter_)
            else:
                formatted_data = ""

            if filter_:
                reset_notify = False
                if f_notify is not False and not filter_.f_notify:
                    reset_notify = True
                    filter_.f_notify = self.lazy_filter_notify

                try:
                    formatted_data = filter_(formatted_data)
                finally:
                    filter_.f_notify = None

            tmp.fnp_got = fnp_got = self._lazy_add_extension(
                self.lazy_fnp_got_root(control=control), extension, suffix
            )
            self._lazy_write(fnp_got, formatted_data)

            # the caller requested not to check exp == got.  this can be done, for example
            # to use the parsing and formatting mechanisms without actually comparing.
            if no_assert:
                if self.verbose:
                    logger.info("%s no_assert:%1.  returning" % (smsg, no_assert))
                return self.lazytemp

            try:
                if self.verbose:
                    logger.info("%s.assertLazy.reading:%s" % (self, fnp_exp))

                with codecs.open(fnp_exp, encoding="utf-8", errors="ignore") as fi:
                    exp = fi.read().strip()
            except (IOError,) as e:
                return control.handler_io_error(fnp_exp, formatted_data, message, exc=e)
            except (Exception,) as e:  # pragma : no cover
                pdb.set_trace()
                raise

            if self.verbose and self.verbose >= 2:
                msg = "\nexp:%s:\n<>\ngot:%s:" % (exp, formatted_data)
                logger.info(msg)

            if control.nodiff:
                message = " not equal but diffing disabled"

            if self.lazy_message_formatter and not message:
                if exp != formatted_data:
                    # pdb.set_trace()
                    if not control.baseline:
                        message = self.lazy_message_formatter.format(
                            exp, formatted_data, window=5
                        )
                    else:
                        message = "no match but baseline mode: will reset %(fnp_exp)s and pass test"

                    if self.verbose:
                        message += (
                            "\n exp @ %(fnp_exp)s \n got @ %(fnp_got)s" % locals()
                        )

                    if rpdb() and not message:
                        pdb.set_trace()

            if control.baseline:
                try:
                    # don't try a comparison, because those that often runs too long
                    # and you're not learning that much anyway on baseline
                    self.assertTrue(exp == formatted_data, message)
                    # self.assertEqual(exp, formatted_data, message)
                except (AssertionError,) as e:
                    e.lazytemp = self.lazytemp
                    self._lazy_write(fnp_exp, formatted_data)
                    logger.warning(u"lazy: %s.  expectation has been reset" % (e))

                return self.lazytemp

            if isinstance(formatted_data, basestring_):
                formatted_data = formatted_data.strip()
            self.assertEqual(exp, formatted_data, message)

            return self.lazytemp

        except (IOError, AssertionError) as e:
            e.lazytemp = self.lazytemp
            raise
        except (Exception,) as e:  # pragma : no cover
            e.lazytemp = getattr(self, "lazytemp", None)

            if cpdb():
                pdb.set_trace()
            raise

    def lazy_debug(self):
        if self.lazytemp:
            print("lazy_debug")
            # logger.info(debugObject(self.lazytemp, "temp"))
            logger.info(debugObject(self.lazytemp.env, "env"))
            logger.info(
                "handler_io_error:%s.%s"
                % (
                    self.lazytemp.control.handler_io_error.__module__,
                    self.lazytemp.control.handler_io_error.func_name,
                )
            )


def output_help():

    if getattr("output_help", "done", False):
        return

    writer = sys.stderr.write

    writer("\n")
    writer("*" * 80)
    writer("\n")

    # pdb.set_trace()

    module = "lazy-regression-tests"

    writer(
        "%s behavior can be controlled with flags or by environment variable $lzrt_directive\n"
        % (module)
    )

    writer("available flags:\n")
    writer(
        "- %s establishes baseline behavior - IOError and mismatches pass\n"
        % (SYSARG_BASELINE)
    )
    writer("- %s don't run regression tests\n" % (SYSARG_IGNORE))
    writer("- %s pass tests with missing expectations\n" % (SYSARG_PASS_MISSING))
    writer("- %s check equality but dont run diff s\n" % (SYSARG_NODIFF))

    DirectiveChoices.output_help(writer)

    writer("\n")
    writer("*" * 80)

    output_help.done = True


if "-h" in sys.argv:
    output_help()


def lazy_pass_missing(*classes):
    # if "-h" in sys.argv:
    #     output_help()

    if SYSARG_PASS_MISSING in sys.argv:

        for cls_ in classes:
            cls_.lazy_environ[env_directive] = OnAssertionError.pass_missing
        sys.argv.remove(SYSARG_PASS_MISSING)


def lazy_nodiff(*classes):
    # if "-h" in sys.argv:
    #     output_help()
    # pdb.set_trace()

    if SYSARG_NODIFF in sys.argv:

        for cls_ in classes:
            cls_.lazy_environ[env_directive] = OnAssertionError.nodiff
        sys.argv.remove(SYSARG_NODIFF)


def lazy_baseline(*classes):
    # if "-h" in sys.argv:
    #     output_help()

    if SYSARG_BASELINE in sys.argv:

        for cls_ in classes:
            cls_.lazy_environ[env_directive] = OnAssertionError.baseline
        sys.argv.remove(SYSARG_BASELINE)


def lazy_ignore(*classes):
    # if "-h" in sys.argv:
    #     output_help()

    if SYSARG_IGNORE in sys.argv:

        for cls_ in classes:
            cls_.lazy_environ[env_directive] = OnAssertionError.ignore
        sys.argv.remove(SYSARG_IGNORE)



from lazy_regression_tests._baseutils import set_cpdb, set_rpdb, ppp, debugObject

import pdb
from bemyerp.lib.utils import set_cpdb, set_rpdb


def cpdb(**kwds) -> bool:  # pragma : no cover
    if cpdb.enabled == "once":
        cpdb.enabled = False  # type : ignore
        return True
    return cpdb.enabled  # type : ignore


cpdb.enabled = False  # type : ignore


from lazy_regression_tests.utils import (
    _Filter,
    RemoveTextFilter,
    DataMatcher,
    ppp,
    fill_template,
    Subber,
    RescueDict,
    NamesMatchTemp,
)


class FormatterRemoveFilter(DataMatcher):
    raw_format_filter = True


class CSSRemoveFilter(FormatterRemoveFilter):
    def __init__(self, css, *args, **kwds):
        self.css = css
        self.verbose = kwds.get("verbose")
        self.hitname = kwds.get("hitname")
        self.temp = NamesMatchTemp()

    def format(self, soup, lazytemp, **kwds):

        try:
            for hit in soup.select(self.css):

                if self.hitname:
                    lazytemp.notify(repr(hit), self)

                hit.decompose()

            return soup
        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    __call__ = format



import re
import os
import difflib
from collections import deque, namedtuple
from functools import partial


import pdb

from lazy_regression_tests._baseutils import (
    set_cpdb,
    set_rpdb,
    ppp,
    debugObject,
    fill_template,
    Subber,
    RescueDict,
)

from traceback import print_exc as xp

import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


###################################################################
# Python 2 to 3.  !!!TODO!!!p4- Simplify after Python support ends.
###################################################################
try:
    basestring_ = basestring
except (NameError,) as e:
    basestring_ = str

try:
    unicode_ = unicode
except (NameError,) as e:
    unicode_ = str


def cpdb(**kwds):  # pragma : no cover
    if cpdb.enabled == "once":
        cpdb.enabled = False  # type: ignore
        return True
    return cpdb.enabled


cpdb.enabled = False  # type: ignore


def rpdb():  # pragma : no cover
    return rpdb.enabled


rpdb.enabled = False  # type: ignore
Found = namedtuple("Found", "found by")


def found_repr(self):
    return "found:%-30.30s: by:%s:" % (self[0], self[1])


Found.__repr__ = found_repr


class DiffFormatter(object):
    """inherit from object to make it a new-style class, else super will complain"""

    def __init__(self, *args, **kwds):

        self.window = kwds.pop("window", None)
        self.maxlines = kwds.pop("maxlines", None)

        self._differ = difflib.Differ()

        # super(DiffFormatter, self).__init__(self, *args, **kwds)

    def _window(self, lines, window=None):
        try:

            if not window:
                return lines
            if not isinstance(window, int):
                raise TypeError("window has to be an int")

            # remember, at most, window # of lines
            dq = deque(maxlen=window)
            cntr = 0

            res = []

            for line in lines:

                if line[0] in ("+", "-"):
                    # cntr, while > 0 adds line to res
                    cntr = window
                    while True:
                        try:
                            # try if res.extend(dq) works
                            res.append(dq.popleft())
                        except (IndexError,) as e:
                            break
                    res.append(line)
                elif cntr > 0:
                    cntr -= 1
                    res.append(line)
                else:
                    # this line won't be used, unless a later line
                    # requires it in context.
                    dq.append(line)
            return res
        except (Exception,) as e:  # pragma : no cover
            raise

    def format(self, exp, got, window=None):
        try:
            exp_ = exp.splitlines()
            got_ = got.splitlines()
            lines = self._differ.compare(exp_, got_)

            window = window or self.window

            if window:
                lines2 = self._window(lines, window)
            else:
                lines2 = list(lines)

            if self.maxlines:
                # this doesn't work well 0 1 2 3 ... vs 100 101 102 103 ...
                # will show all the - in the maxlines since there is nothing in common...
                lines2 = lines2[: self.maxlines]

            msg = "\n".join(lines2)
            msg = msg.strip()
            if msg and msg[1] != " ":
                msg = "  %s" % (msg)
            return msg

        except (Exception,) as e:  # pragma : no cover
            raise


def replace(list_, item=None, with_=[]):
    """utility function:
       given an item, if it's found in the list_, replace it there,
       except that lists are expanded at that position
    """

    try:
        try:
            pos = list_.index(item)
        except (ValueError,) as e:
            return list_[:]

        if isinstance(with_, list):
            li = list_[0:pos] + with_ + list_[pos + 1 :]
            return li

        res = list_[:]
        res[pos] = with_
        return res

    except (Exception,) as e:  # pragma : no cover
        if cpdb():
            pdb.set_trace()
        raise


class MediatedEnvironDict(dict):
    def __init__(self, filters=[], seed={}, *args, **kwds):
        """
        :param filters: filter.search(envname) will  
        :param seed   : default values which block env variables
        :param **kwds : passed on to std `dict` constructor
        """

        super(MediatedEnvironDict, self).__init__(*args, **kwds)
        self.set_filters(filters)
        self.acquired = False
        self.seed = seed
        self.update(**self.seed)

    def set_filters(self, filters):
        li = self.filters = []
        for filter_ in filters:
            if isinstance(filter_, basestring_):
                li.append(re.compile(filter_))
            else:
                li.append(filter_)

    def acquire(self):
        """
        read environment variables matching the filters
        note `seed` keys will not get updated"""

        try:

            for k, v in os.environ.items():
                for filter_ in self.filters:
                    if filter_.search(k):
                        # dont override what's already set
                        if not k in self:
                            self[k] = v
                            continue

            self.acquired = True
            # pdb.set_trace()

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    def clear(self):
        """flushes the dict and reapplies the seed"""
        self.acquired = False
        super(MediatedEnvironDict, self).clear()
        self.update(**self.seed)


class Finder(object):

    target = "?"

    def __repr__(self):
        return self.target

    def is_match(self, notify, *names):
        try:

            if len(names) > 1:
                raise NotImplementedError()

            if names[0] == self.target:
                notify(Found(names, self))
        except (Exception,) as e:  # pragma : no cover
            ppp(self, "%s.is_match" % (self))
            logger.info("names:%s" % names)
            if cpdb():
                pdb.set_trace()
            raise

    def __init__(self, target):
        self.target = target
        self._init()

    def _init(self):
        pass


class NamesMatchTemp(object):
    def __init__(self):
        self.finds = []
        self.work = []

    def __repr__(self):
        return """
  hits:%s
  work:%s\n""" % (
            self.finds,
            self.work,
        )


class WorkerNamesMatch(object):
    def __init__(self):
        self.done = {}

    def process(self, target, matchtemp):
        self.done = {}
        self._process(target, matchtemp)

    def _process(self, target, matchtemp):
        raise NotImplementedError("%s.process" % (self))


class RemoveWorkerNamesMatch(WorkerNamesMatch):
    def _process(self, target, matchtemp):
        try:
            di = self.done
            for found in matchtemp.finds:

                if len(found.found) > 1:
                    raise NotImplementedError("%s._process" % (self))

                key = found.found[0]
                di[key] = target.pop(key)

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise


class _Filter(object):
    """hacky because this is different for dictionary than text filtering"""

    def __init__(self, matchers_=[], notify=None, callback=None):

        self.matchers = []
        for matcher in matchers_:
            if isinstance(matcher, Finder):
                self.matchers.append(matcher)
            else:
                self.matchers.append(Finder(matcher))
        self.notify = notify or self._notify
        self.callback = callback

    # def scan(self, target, temp=None):
    def scan(self, target):
        # self.temp = temp or getattr(self, "temp", None) or NamesMatchTemp()
        self.temp = NamesMatchTemp()

        for key, value in target.items():
            #!!!TODO!!! - recurse into nested dicts.

            for matcher in self.matchers:
                if matcher.is_match(self.notify, key):
                    continue

            # recursion error....
            # if isinstance(value, dict):
            #     self.scan(target, self.temp)

        if self.callback:
            self.callback(target, self.temp)
        return self.temp

    def _notify(self, found):
        self.temp.finds.append(found)


NamesMatcher = _Filter


class DictionaryKeyFilter(_Filter):
    def __init__(self, *args, **kwds):
        super(DictionaryKeyFilter, self).__init__(*args, **kwds)
        self.worker = RemoveWorkerNamesMatch()
        self.callback = self.worker.process


class DataMatcher(object):
    def __repr__(self):
        return "%s.%s" % (self.__module__, self.__class__.__name__)

    hitname = None

    raw_format_filter = False

    def __init__(self, *args, **kwds):
        self.verbose = kwds.get("verbose")
        self.hitname = kwds.get("hitname")


class RegexMatcher(DataMatcher):
    def __init__(self, pattern, *args, **kwds):
        self.patre = re.compile(pattern, *args)
        super(RegexMatcher, self).__init__(*args, **kwds)

    def search(self, *args, **kwds):
        return self.patre.search(*args, **kwds)

    def __getattr__(self, attrname):
        return getattr(self.patre, attrname)


class RegexRemoveSaver(RegexMatcher):
    """this will remove the matching line but also save it"""

    def __init__(self, pattern, *args, **kwds):
        assert kwds.get("hitname"), "RegexRemoveSaver needs a hitname"
        super(RegexRemoveSaver, self).__init__(pattern, *args, **kwds)


class RegexSubstitHardcoded(RegexMatcher):
    """allows for replacement of the line with different contents

       can't use a re.sub directly because the Filter won't know if 
       it's just a match filter or a match & substitution
    """

    def __repr__(self):

        subinfo = None

        try:

            if callable(self.substitution):
                if isinstance(self.substitution, partial):
                    subinfo = "partial:%s" % self.substitution.func.__name__
                else:
                    subinfo = "func.%s" % self.substitution.__name__
            else:
                subinfo = str(self.substitution)
        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

        return "%s.(pattern=%s, substitution=%s" % (
            self.__class__.__name__,
            self.pattern,
            subinfo,
        )

    def __init__(self, pattern, substitution, *args):

        self.patre = re.compile(pattern, *args)
        self.substitution = substitution

    def __getattr__(self, attrname):
        return getattr(self.patre, attrname)

    def substitute(self, line):
        return self.substitution


class RegexSubstitFilter(RegexSubstitHardcoded):
    def __init__(self, pattern, substitution, *args, **kwds):

        self.verbose = kwds.get("verbose")

        if not callable(substitution):
            raise TypeError(
                "substitution needs to be a callable f(re.MatchObject)=>str"
            )

        super(RegexSubstitFilter, self).__init__(pattern, substitution, *args)

    def substitute(self, line):
        try:
            res = self.patre.sub(self.substitution, line)
            # if rpdb(): # pragma : no cover
            #     pdb.set_trace()
            if self.verbose:
                logger.info("\n  %s\n  :%s:\n  =>\n  :%s:" % (self, line, res))
            return res

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise


class KeepTextFilter(object):

    KEEP = True

    dflt_cls = RegexMatcher

    def __init__(self, regexes=[], f_notify=None):

        """:param regexes: list of regex's.  or strings which will be compiled to regex
           you could also pass your own matching objects, they need `search(string)=>boolean`
           method
        """

        try:
            self.formatters = []

            regexes_ = regexes[:]

            regexes_ = []
            for regex in regexes:
                if getattr(regex, "raw_format_filter", False):
                    self.formatters.append(regex)
                    continue

                if not isinstance(regex, DataMatcher):
                    regex = self.dflt_cls(re.compile(regex))
                regexes_.append(regex)

            self.regexes = regexes_
            self.f_notify = f_notify

        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    def copy(self):
        """return a duplicate, but with a copy of the regexes..."""

        newinst = self.__class__(regexes=[], f_notify=self.f_notify)
        newinst.regexes = self.regexes[:]
        return newinst

    def add_regex(self, regex):
        if not isinstance(regex, self.dflt_cls):
            regex = self.dflt_cls(re.compile(regex))
        self.regexes.append(regex)

    def _is_match(self, line):
        try:
            res = False
            for regex in self.regexes:
                if regex.search(line):
                    if self.f_notify:
                        self.f_notify(Found(line, regex))

                    # these are special classes such as
                    # utils.RegexSubstitHardcoded
                    # utils.RegexSubstitFilter
                    substitute = getattr(regex, "substitute", None)
                    if substitute:
                        line = substitute(line)
                        # but if we're called from RemoveTextFilter
                        # then keep the line
                        return self.KEEP, line

                    return True, line
            return False, line
        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise

    def filter(self, formatted_data):

        if not isinstance(formatted_data, list):
            formatted_data = formatted_data.splitlines()

        lines = []
        for line in formatted_data:
            keep, line = self._is_match(line)
            if keep:
                lines.append(line)
        return "\n".join(lines)

    __call__ = filter

    def format(self, data, lazytemp, **kwargs):

        if not self.formatters:
            return data

        try:
            for formatter in self.formatters:
                data = formatter.format(data, lazytemp=lazytemp, **kwargs)

            return data

            # raise NotImplementedError("format(%s)" % (locals()))
        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            raise


class RemoveTextFilter(KeepTextFilter):

    KEEP = False

    def filter(self, formatted_data):
        lines = []
        for line in formatted_data.splitlines():
            keep, line = self._is_match(line)
            if not keep:
                lines.append(line)
        return "\n".join(lines)

    __call__ = filter

    # def format(self, data, **kwargs):

    #     if not self.formatters:
    #         return data

    #     try:
    #         raise NotImplementedError("format(%s)" % (locals()))
    #     except (Exception,) as e: # pragma : no cover
    #         if cpdb(): pdb.set_trace()
    #         raise


def curry_func(func, replacement, verbose=False, **kwds):
    """preps the function for replacement"""
    return partial(func, replacement=replacement, verbose=verbose, **kwds)


def simple_subber(match, *args, **kwds):

    """
        meant to be called as per re.sub(pattern, callable)
        :param match:  is a re.MatchObject
        :param replacement: keyword parameter - required - string to replace with
               use `operator.partial` or `curry_func` 

               ex: `partial(simple_subber, replacement="myreplacement"))`

        :param onlyonce: keyword parameter - used to avoid replacing on a reformat when 
        the replacement is found in the match string.
        :return: a string as per result of re.sub(pattern, callable)
    """

    try:
        try:
            replacement = kwds["replacement"]
        except (KeyError,) as e:
            raise TypeError(
                "missing replacement keyword parameter.  refer to docstring"
            )
            raise

        rpdb = kwds.get("rpdb")

        verbose = kwds.get("verbose")

        # avoid re doing multiple replaces...
        onlyonce = kwds.get("onlyonce")
        if onlyonce and replacement in match.string:
            return match.string[match.start() : match.end()]

        assert isinstance(replacement, basestring_)
        if rpdb or verbose:
            # ppp(match)

            di = {
                "match.re.pattern": match.re.pattern,
                "match.groups": match.groups(),
                "match.string": match.string,
            }
            ppp(di)

        try:
            res = match.string[match.start() : match.end()].replace(
                match.groups(0)[0], replacement
            )
        except (Exception,) as e:  # pragma : no cover
            if cpdb():
                pdb.set_trace()
            return "!lazy_regression_tests:bad substitution:!"

        if verbose:
            print("\nsubstit02:\n  %s \n  =>\n  %s" % (match.string, res))
        if rpdb:
            pdb.set_trace()

        # raise NotImplementedError()
        return res

    except (Exception,) as e:  # pragma : no cover
        if cpdb():
            pdb.set_trace()
        raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = []

setup_requirements = []

test_requirements = []

setup(
    author="JL Peyret",
    author_email="jpeyret@gmail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        # "Programming Language :: Python :: 2",      could come back, but working on mypy in v3 for now
        # "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    description="a very lazy way to regression test almost anything that has consistent structured text output",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="lazy_regression_tests",
    long_description_content_type="text/markdown",
    name="lazy_regression_tests",
    packages=find_packages(include=["lazy_regression_tests"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/jpeyret/lazy_regression_tests",
    version="0.2.0",
    zip_safe=False,
)
